
/*
 *  parse.c,v 1.66 1994/01/24 18:32:00 jch Exp
 */

/* Gated Release 3.1 */
/* Copyright (c) 1990,1991,1992,1993 by Cornell University. All rights reserved. */
/* Refer to Particulars and other Copyright notices at the end of this file. */


#define	INCLUDE_DIRENT
#define	INCLUDE_ROUTE
#include "include.h"
#include "parse.h"
#include "krt.h"
#ifdef	PROTO_RIP
#include "rip.h"
#endif	/* PROTO_RIP */
#ifdef	PROTO_HELLO
#include "hello.h"
#endif	/* PROTO_HELLO */
#ifdef	PROTO_OSPF
#include "ospf.h"
#endif	/* PROTO_OSPF */
#ifdef	PROTO_EGP
#include "egp.h"
#endif	/* PROTO_EGP */
#ifdef	PROTO_BGP
#include "bgp.h"
#endif	/* PROTO_BGP */
#ifdef	PROTO_SLSP
#include "slsp.h"
#endif	/* PROTO_SLSP */
#ifdef	PROTO_ISIS
#include "isis.h"
#endif	/* PROTO_ISIS */
#ifdef	PROTO_DVMRP
#include "dvmrp.h"
#endif	/* PROTO_DVMRP */
#include "parser.h"

static int n_keywords;

/*
 *	Table of keywords recognized.  This table is sorted at runtime
 *	to facilitate binary searching.
 */

static bits keywords[] =
{
/* Start Sort */
#ifdef	TISIS_BUILDLSP
    {TISIS_BUILDLSP, "buildlsp"},
#endif	/* TISIS_BUILDLSP */
#ifdef	TISIS_CSNP
    {TISIS_CSNP, "csnp"},
#endif	/* TISIS_CSNP */
#ifdef	TISIS_DUMPLSP
    {TISIS_DUMPLSP, "dumplsp"},
#endif	/* TISIS_DUMPLSP */
#ifdef	TISIS_EVENTS
    {TISIS_EVENTS, "events"},
#endif	/* TISIS_EVENTS */
#ifdef	TISIS_FLOODING
    {TISIS_FLOODING, "flooding"},
#endif	/* TISIS_FLOODING */
#ifdef	TISIS_IIH
    {TISIS_IIH, "iih"},
#endif	/* TISIS_IIH */
#ifdef	TISIS_LANADJ
    {TISIS_LANADJ, "lanadj"},
#endif	/* TISIS_LANADJ */
#ifdef	TISIS_LSPCONTENT
    {TISIS_LSPCONTENT, "lspcontent"},
#endif	/* TISIS_LSPCONTENT */
#ifdef	TISIS_LSPDB
    {TISIS_LSPDB, "lspdb"},
#endif	/* TISIS_LSPDB */
#ifdef	TISIS_LSPINPUT
    {TISIS_LSPINPUT, "lspinput"},
#endif	/* TISIS_LSPINPUT */
#ifdef	TISIS_P2PADJ
    {TISIS_P2PADJ, "p2padj"},
#endif	/* TISIS_P2PADJ */
#ifdef	TISIS_PATHS
    {TISIS_PATHS, "paths"},
#endif	/* TISIS_PATHS */
#ifdef	TISIS_PSNP
    {TISIS_PSNP, "psnp"},
#endif	/* TISIS_PSNP */
#ifdef	TISIS_SUMMARY
    {TISIS_SUMMARY, "summary"},
#endif	/* TISIS_SUMMARY */
#ifdef	T_ACK
    {T_ACK, "ack"},
    {T_ACK, "acknowledgement"},
#endif	/* T_ACK */
#ifdef	T_ACQUIRE
    {T_ACQUIRE, "acquire"},
#endif	/* T_ACQUIRE */
#ifdef	T_ADV
    {T_ADV,	"adv"},
    {T_ADV,	"adv_entry"},
#endif	/* T_ADV */
#ifdef	T_AGGREGATE
    {T_AGGREGATE, "aggregate"},
#endif	/* T_AGGREGATE */
#ifdef	T_ALERT
    {T_ALERT, "alert"},
#endif	/* T_ALERT */
#ifdef	T_ALL
    {T_ALL, "all"},
#endif	/* T_ALL */
#ifdef	T_ALLOW
    {T_ALLOW, "allow" },
#endif	/* T_ALLOW */
#ifdef	T_ANALRETENTIVE
    {T_ANALRETENTIVE, "anal-retentive" },
    {T_ANALRETENTIVE, "analretentive" },
#endif	/* T_ANALRETENTIVE */
#ifdef	T_ANY
    {T_ANY,	"any"},
#endif	/* T_ANY */
#ifdef	T_AREA
    {T_AREA, "area"},
#endif	/* T_AREA */
#ifdef	T_AS
    {T_AS, "as"},
    {T_AS, "autonomous-system"},
    {T_AS, "autonomoussystem"},
#endif	/* T_AS */
#ifdef	T_ASPATH
    {T_ASPATH,	"as-path"},
    {T_ASPATH,	"aspath"},
#endif	/* T_ASPATH */
#ifdef	T_AUTH
    {T_AUTH, "auth"},
    {T_AUTH, "authentication"},
#endif	/* T_AUTH */
#ifdef	T_AUTHKEY
    {T_AUTHKEY, "auth-key"},
    {T_AUTHKEY, "authentication-key"},
    {T_AUTHKEY, "authenticationkey"},
    {T_AUTHKEY, "authkey"},
#endif	/* T_AUTHKEY */
#ifdef	T_AUTHTYPE
    {T_AUTHTYPE, "auth-type"},
    {T_AUTHTYPE, "authentication-type"},
    {T_AUTHTYPE, "authenticationtype"},
    {T_AUTHTYPE, "authtype"},
#endif	/* T_AUTHTYPE */
#ifdef	T_BACKBONE
    {T_BACKBONE, "backbone"},
#endif	/* T_BACKBONE */
#ifdef	T_BACKGROUND
    {T_BACKGROUND, "background"},
#endif	/* T_BACKGROUND */
#ifdef	T_BGP
    {T_BGP, "bgp"},
#endif	/* T_BGP */
#ifdef	T_BLACKHOLE
    {T_BLACKHOLE, "blackhole"},
#endif	/* T_BLACKHOLE */
#ifdef	T_BRIEF
    {T_BRIEF, "brief"},
#endif	/* T_BRIEF */
#ifdef	T_BROADCAST
    {T_BROADCAST, "broadcast"},
#endif	/* T_BROADCAST */
#ifdef	T_CIRCUIT
    {T_CIRCUIT,	"circuit"},
#endif	/* T_CIRCUIT */
#ifdef	T_CRIT
    {T_CRIT, "crit"},
#endif	/* T_CRIT */
#ifdef	T_DD
    {T_DD, "dd"},
#endif	/* T_DD */
#ifdef	T_DEBUG
    {T_DEBUG, "debug"},
#endif	/* T_DEBUG */
#ifdef	T_DEFAULT
    {T_DEFAULT, "default"},
#endif	/* T_DEFAULT */
#ifdef	T_DEFAULTIN
    {T_DEFAULTIN,	"importdefault"},
#endif	/* T_DEFAULTIN */
#ifdef	T_DEFAULTMETRIC
    {T_DEFAULTMETRIC, "defaultmetric"},
#endif	/* T_DEFAULTMETRIC */
#ifdef	T_DEFAULTOUT
    {T_DEFAULTOUT,	"exportdefault"},
#endif	/* T_DEFAULTOUT */
#ifdef	T_DEFAULTS
    {T_DEFAULTS, "defaults"},
#endif	/* T_DEFAULTS */
#ifdef	T_DEFINE
    {T_DEFINE,		"define" },
#endif	/* T_DEFINE */
#ifdef	T_DETAIL
    {T_DETAIL, "detail"},
#endif	/* T_DETAIL */
#ifdef	T_DIRECT
    {T_DIRECT, "direct"},
#endif	/* T_DIRECT */
#ifdef	T_DISABLE
    {T_DISABLE, "disable"},
    {T_DISABLE, "disabled"},
#endif	/* T_DISABLE */
#ifdef	T_DOWN
    {T_DOWN, "down"},
#endif	/* T_DOWN */
#ifdef	T_DUAL
    {T_DUAL,	"dual"},
#endif	/* T_DUAL */
#ifdef	T_DVMRP
    {T_DVMRP, "dvmrp"},
#endif	/* T_DVMRP */
#ifdef	T_EGP
    {T_EGP, "egp"},
#endif	/* T_EGP */
#ifdef	T_ELIGIBLE
    {T_ELIGIBLE, "eligible"},
#endif	/* T_ELIGIBLE */
#ifdef	T_EMERG
    {T_EMERG, "emerg" },
#endif	/* T_EMERG */
#ifdef	T_ENABLE
    {T_ENABLE, "enable"},
    {T_ENABLE, "enabled"},
#endif	/* T_ENABLE */
#ifdef	T_ERR
    {T_ERR, "err"},
#endif	/* T_ERR */
#ifdef	T_ERROR
    {T_ERROR, "error"},
#endif	/* T_ERROR */
#ifdef	T_EVERY
    {T_EVERY, "every"},
#endif	/* T_EVERY */
#ifdef	T_EXACT
    {T_EXACT, "exact"},
#endif	/* T_EXACT */
#ifdef	T_EXCEPT
    {T_EXCEPT, "except"},
#endif	/* T_EXCEPT */
#ifdef	T_EXPORT
    {T_EXPORT, "export"},
#endif	/* T_EXPORT */
#ifdef	T_EXPORTINTERVAL
    {T_EXPORTINTERVAL, "export-interval"},
    {T_EXPORTINTERVAL, "exportinterval"},
#endif	/* T_EXPORTINTERVAL */
#ifdef	T_EXPORTLIMIT
    {T_EXPORTLIMIT, "export-limit"},
    {T_EXPORTLIMIT, "exportlimit"},
#endif	/* T_EXPORTLIMIT */
#ifdef	T_EXTERNAL
    {T_EXTERNAL, "external"},
#endif	/* T_EXTERNAL */
#ifdef	T_FILES
    {T_FILES, "files"},
#endif	/* T_FILES */
#ifdef	T_FIRST
    {T_FIRST, "first"},
#endif	/* T_FIRST */
#ifdef	T_FLASH
     {T_FLASH, "flash"},
#endif	/* T_FLASH */
#ifdef	T_GATEWAY
    {T_GATEWAY, "gateway"},
    {T_GATEWAY, "gw"},
    {T_GATEWAY, "router"},
    {T_GATEWAY, "rtr"},
#endif	/* T_GATEWAY */
#ifdef	T_GENDEFAULT
    {T_GENDEFAULT, "gendefault"},
    {T_GENDEFAULT, "originate-default"},
    {T_GENDEFAULT, "originatedefault"},
#endif	/* T_GENDEFAULT */
#ifdef	T_GENERAL
    {T_GENERAL, "general"},
#endif	/* T_GENERAL */
#ifdef	T_GENERATE
    {T_GENERATE,	"generate"},
#endif	/* T_GENERATE */
#ifdef	T_GROUP
    {T_GROUP, "group"},
#endif	/* T_GROUP */
#ifdef	T_HELLO
    {T_HELLO, "hello"},
#endif	/* T_HELLO */
#ifdef	T_HELLOINTERVAL
    {T_HELLOINTERVAL, "hello-interval"},
    {T_HELLOINTERVAL, "hellointerval"},
#endif	/* T_HELLOINTERVAL */
#ifdef	T_HIGHER
    {T_HIGHER, "higher"},
#endif	/* T_HIGHER */
#ifdef	T_HOLDTIME
    {T_HOLDTIME, "holdtime"},
#endif	/* T_HOLDTIME */
#ifdef	T_HOST
    {T_HOST, "host"},
#endif	/* T_HOST */
#ifdef	T_ICMP
    {T_ICMP, "icmp"},
#endif	/* T_ICMP */
#ifdef	T_IDPR
    {T_IDPR, "idpr"},
#endif	/* T_IDPR */
#ifdef	T_IFLIST
    {T_IFLIST, "iflist"},
    {T_IFLIST, "if-list"},
    {T_IFLIST, "interfacelist"},
    {T_IFLIST, "interface-list"},
#endif	/* T_IFLIST */
#ifdef	T_IGMP
    {T_IGMP, "igmp"},
#endif	/* T_IGMP */
#ifdef	T_IGNORE
    {T_IGNORE, "ignore"},
#endif	/* T_IGNORE */
#ifdef	T_IGP
    {T_IGP,	"igp"},
#endif	/* T_IGP */
#ifdef	T_IMPORT
    {T_IMPORT, "import"},
#endif	/* T_IMPORT */
#ifdef	T_INCOMPLETE
    {T_INCOMPLETE, "incomplete"},
#endif	/* T_INCOMPLETE */
#ifdef	T_INDELAY
    {T_INDELAY, "indelay"},
    {T_INDELAY, "in-delay"},
#endif	/* T_INDELAY */
#ifdef	T_INET
    {T_INET,	"inet"},
#endif	/* T_INET */
#ifdef	T_INFINITY
    {T_INFINITY, "infinity"},
    {T_INFINITY, "unreachable"},
#endif	/* T_INFINITY */
#ifdef	T_INFO
    {T_INFO, "info"},
#endif	/* T_INFO */
#ifdef	T_INFTRANSDELAY
    {T_INFTRANSDELAY, "inftransdelay"},
    {T_INFTRANSDELAY, "transit-delay"},
    {T_INFTRANSDELAY, "transitdelay"},
#endif	/* T_INFTRANSDELAY */
#ifdef	T_INTDOMINFO
    {T_INTDOMINFO,	"internal-domain-info"},
    {T_INTDOMINFO,	"internaldomaininfo"},
#endif	/* T_INTDOMINFO */
#ifdef	T_INTERFACE
    {T_INTERFACE, "interface"},
    {T_INTERFACE, "intf"},
#endif	/* T_INTERFACE */
#ifdef	T_INTERFACES
    {T_INTERFACES, "interfaces"},
    {T_INTERFACES, "intfs"},
#endif	/* T_INTERFACES */
#ifdef	T_INTERIOR
    {T_INTERIOR, "interior"},
#endif	/* T_INTERIOR */
#ifdef	T_INTERNAL
    {T_INTERNAL, "internal"},
#endif	/* T_INTERNAL */
#ifdef	T_IP
    {T_IP,	"ip"},
#endif	/* T_IP */
#ifdef	T_IPREACH
    {T_IPREACH,	"ip-reachability"},
    {T_IPREACH,	"ipreachability"},
#endif	/* T_IPREACH */
#ifdef	T_ISIS
    {T_ISIS,	"is-is"},
    {T_ISIS,	"isis"},
#endif	/* T_ISIS */
#ifdef	T_ISO
    {T_ISO,	"iso"},
#endif	/* T_ISO */
#ifdef	T_K
    {T_K, "k"},
#endif	/* T_K */
#ifdef	T_KEEPALIVE
    {T_KEEPALIVE, "keepalive"},
#endif	/* T_KEEPALIVE */
#ifdef	T_KEEP
    {T_KEEP, "keep"},
    {T_KEEP, "keep"},
#endif	/* T_KEEP */
#ifdef	T_KERNEL
    {T_KERNEL, "kernel"},
#endif	/* T_KERNEL */
#ifdef	T_LCLADDR
    {T_LCLADDR, "lcl-addr"},
    {T_LCLADDR, "lcladdr"},
    {T_LCLADDR, "local-address"},
    {T_LCLADDR, "localaddress"},
#endif	/* T_LCLADDR */
#ifdef	T_LEVEL
    {T_LEVEL,	"level"},
#endif	/* T_LEVEL */
#ifdef	T_LIMIT
     {T_LIMIT, "limit"},
#endif	/* T_LIMIT */
#ifdef	T_LOCALAS
    {T_LOCALAS, "asout"},
    {T_LOCALAS, "localas" },
#endif	/* T_LOCALAS */
#ifdef	T_LOOPS
    {T_LOOPS, "loops"},
#endif	/* T_LOOPS */
#ifdef	T_LOWER
    {T_LOWER, "lower"},
#endif	/* T_LOWER */
#ifdef	T_LSA_BLD
    {T_LSA_BLD, "lsa-build"},
    {T_LSA_BLD, "lsabuild"},
#endif	/* T_LSA_BLD */
#ifdef	T_LSA_RX
    {T_LSA_RX, "lsa-receive"},
    {T_LSA_RX, "lsa-rx"},
    {T_LSA_RX, "lsareceive"},
    {T_LSA_RX, "lsarx"},
#endif	/* T_LSA_RX */
#ifdef	T_LSA_TX
    {T_LSA_TX, "lsa-transmit"},
    {T_LSA_TX, "lsa-tx"},
    {T_LSA_TX, "lsatransmit"},
    {T_LSA_TX, "lsatx"},
#endif	/* T_LSA_TX */
#ifdef	T_M
    {T_M, "m"},
#endif	/* T_M */
#ifdef	T_MARK
    {T_MARK, "mark"},
#endif	/* T_MARK */
#ifdef	T_MARTIANS
    {T_MARTIANS, "martians"},
#endif	/* T_MARTIANS */
#ifdef	T_MASK
    {T_MASK, "mask"},
#endif	/* T_MASK */
#ifdef	T_MASKLEN
    {T_MASKLEN, "mask-len"},
    {T_MASKLEN, "mask-length"},
    {T_MASKLEN, "masklen"},
    {T_MASKLEN, "masklength"},
#endif	/* T_MASKLEN */
#ifdef	T_MAXUP
    {T_MAXUP, "maxup"},
#endif	/* T_MAXUP */
#ifdef	T_METRIC
    {T_METRIC, "cost"},
    {T_METRIC, "distance"},
    {T_METRIC, "metric"},
#endif	/* T_METRIC */
#ifdef	T_METRICIN
    {T_METRICIN, "metricin"},
#endif	/* T_METRICIN */
#ifdef	T_METRICOUT
    {T_METRICOUT, "metricout"},
#endif	/* T_METRICOUT */
#ifdef	T_MODE
    {T_MODE,	"mode"},
#endif	/* T_MODE */
#ifdef	T_MONITORAUTHKEY
    {T_MONITORAUTHKEY,	"mon-auth-key"},
    {T_MONITORAUTHKEY,	"monauthkey"},
    {T_MONITORAUTHKEY,	"monitor-authentication-key"},
    {T_MONITORAUTHKEY,	"monitorauthenticationkey"},
#endif	/* T_MONITORAUTHKEY */
#ifdef	T_MULTICAST
    {T_MULTICAST,	"multicast" },
#endif	/* T_MULTICAST */
#ifdef	T_NEIGHBOR
    {T_NEIGHBOR, "neighbor"},
    {T_NEIGHBOR, "peer"},
#endif	/* T_NEIGHBOR */
#ifdef	T_NEIGHBORID
    {T_NEIGHBORID, "neighbor-id"},
    {T_NEIGHBORID, "neighborid"},
#endif	/* T_NEIGHBORID */
#ifdef	T_NETMASK
    {T_NETMASK,		"net-mask" },
    {T_NETMASK,		"netmask" },
    {T_NETMASK,		"subnetmask" },
#endif	/* T_NETMASK */
#ifdef	T_NETWORKS
    {T_NETWORKS, "networks"},
#endif	/* T_NETWORKS */
#ifdef	T_NOAGGRID
    {T_NOAGGRID, "noaggregatorid"},
    {T_NOAGGRID, "no-aggregator-id"},
#endif	/* T_NOAGGRID */
#ifdef	T_NOCHECKZERO
    {T_NOCHECKZERO, "no-check-zero"},
    {T_NOCHECKZERO, "nocheckzero"},
#endif	/* T_NOCHECKZERO */
#ifdef	T_NOGENDEFAULT
    {T_NOGENDEFAULT, "nogendefault"},
#endif	/* T_NOGENDEFAULT */
#ifdef	T_NOHELLOIN
    {T_NOHELLOIN, "nohelloin"},
#endif	/* T_NOHELLOIN */
#ifdef	T_NOHELLOOUT
    {T_NOHELLOOUT, "nohelloout"},
#endif	/* T_NOHELLOOUT */
#ifdef	T_NOINSTALL
    {T_NOINSTALL, "noinstall"},
#endif	/* T_NOINSTALL */
#ifdef	T_NONBROADCAST
    {T_NONBROADCAST, "nbma"},
    {T_NONBROADCAST, "no-broadcast"},
    {T_NONBROADCAST, "nobroadcast"},
    {T_NONBROADCAST, "non-broadcast"},
    {T_NONBROADCAST, "nonbroadcast"},
#endif	/* T_NONBROADCAST */
#ifdef	T_NONE
    {T_NONE, "none"},
#endif	/* T_NONE */
#ifdef	T_NOREDIRECTS
    {T_NOREDIRECTS, "noredirects"},
#endif	/* T_NOREDIRECTS */
#ifdef	T_NORESOLV
    {T_NORESOLV, "noresolv"},
    {T_NORESOLV, "noresolve"},
#endif	/* T_NORESOLV */
#ifdef	T_NORIPIN
    {T_NORIPIN, "noripin"},
#endif	/* T_NORIPIN */
#ifdef	T_NORIPOUT
    {T_NORIPOUT, "noripout"},
#endif	/* T_NORIPOUT */
#ifdef	T_NORMAL
    {T_NORMAL, "normal"},
#endif	/* T_NORMAL */
#ifdef	T_NOSEND
    {T_NOSEND, "nosend"},
#endif	/* T_NOSEND */
#ifdef	T_NOSTAMP
    {T_NOSTAMP, "nostamp"},
#endif	/* T_NOSTAMP */
#ifdef	T_NOTICE
    {T_NOTICE, "notice"},
#endif	/* T_NOTICE */
#ifdef	T_OFF
    {T_OFF, "no"},
    {T_OFF, "off"},
#endif	/* T_OFF */
#ifdef	T_ON
    {T_ON, "on"},
    {T_ON, "yes"},
#endif	/* T_ON */
#ifdef	T_OPEN
    {T_OPEN, "open"},
#endif	/* T_OPEN */
#ifdef	T_OPTIONS
    {T_OPTIONS, "options"},
#endif	/* T_OPTIONS */
#ifdef	T_ORIGIN
    {T_ORIGIN,	"origin"},
#endif	/* T_ORIGIN */
#ifdef	T_OSPF
    {T_OSPF, "ospf"},
#endif	/* T_OSPF */
#ifdef	T_OSPF_ASE
    {T_OSPF_ASE, "ospf-ase"},
    {T_OSPF_ASE, "ospfase"},
#endif	/* T_OSPF_ASE */
#ifdef	T_OTHER
    {T_OTHER, "other"},
#endif	/* T_OTHER */
#ifdef	T_OUTDELAY
    {T_OUTDELAY, "outdelay"},
    {T_OUTDELAY, "out-delay"},
#endif	/* T_OUTDELAY */
#ifdef	T_P1
    {T_P1, "minhello"},
    {T_P1, "p1"},
#endif	/* T_P1 */
#ifdef	T_P2
    {T_P2, "minpoll"},
    {T_P2, "p2"},
#endif	/* T_P2 */
#ifdef	T_PACKETS
    {T_PACKETS, "packets"},
#endif	/* T_PACKETS */
#ifdef	T_PARSE
    {T_PARSE, "parse"},
#endif	/* T_PARSE */
#ifdef	T_PASSIVE
    {T_PASSIVE, "passive"},
#endif	/* T_PASSIVE */
#ifdef	T_PEERAS
    {T_PEERAS, "asin"},
    {T_PEERAS, "peeras" },
#endif	/* T_PEERAS */
#ifdef	T_PKTSIZE
    {T_PKTSIZE, "packet-size"},
    {T_PKTSIZE, "packetsize"},
#endif	/* T_PKTSIZE */
#ifdef	T_POINTOPOINT
    {T_POINTOPOINT,	"p2p"},
    {T_POINTOPOINT,	"point2point"},
    {T_POINTOPOINT,	"pointopoint"},
    {T_POINTOPOINT,	"pointtopoint"},
#endif	/* T_POINTOPOINT */
#ifdef	T_POLICY
    {T_POLICY, "policy"},
#endif	/* T_POLICY */
#ifdef	T_POLLINTERVAL
    {T_POLLINTERVAL, "poll-interval"},
    {T_POLLINTERVAL, "pollinterval"},
#endif	/* T_POLLINTERVAL */
#ifdef	T_PORT
    {T_PORT, "port"},
#endif	/* T_PORT */
#ifdef	T_PREFERENCE
    {T_PREFERENCE, "pref"},
    {T_PREFERENCE, "preference"},
#endif	/* T_PREFERENCE */
#ifdef	T_PREFIX
    {T_PREFIX,	"prefix"},
#endif	/* T_PREFIX */
#ifdef	T_PRIORITY
    {T_PRIORITY,	"priority"},
#endif	/* T_PRIORITY */
#ifdef	T_PROTO
    {T_PROTO, "proto"},
    {T_PROTO, "protocol"},
#endif	/* T_PROTO */
#ifdef	T_RECEIVE
    {T_RECEIVE, "receive"},
    {T_RECEIVE, "recv"},
    {T_RECEIVE, "rx"},
#endif	/* T_RECEIVE */
#ifdef	T_RECVBUF
    {T_RECVBUF, "recv-buffer"},
    {T_RECVBUF, "recvbuf"},
    {T_RECVBUF, "recvbuffer"},
#endif	/* T_RECVBUF */
#ifdef	T_REDIRECT
    {T_REDIRECT, "redirect"},
#endif	/* T_REDIRECT */
#ifdef	T_REDIRECTS
    {T_REDIRECTS, "redirects"},
#endif	/* T_REDIRECTS */
#ifdef	T_REFINE
    {T_REFINE, "refine"},
    {T_REFINE, "refines"},
#endif	/* T_REFINE */
#ifdef	T_REGISTER
    {T_REGISTER, "register"},
#endif	/* T_REGISTER */
#ifdef	T_REJECT
    {T_REJECT, "reject"},
#endif	/* T_REJECT */
#ifdef	T_REMNANTS
    {T_REMNANTS, "remnants"},
#endif	/* T_REMNANTS */
#ifdef	T_REPLACE
    {T_REPLACE, "replace"},
#endif	/* T_REPLACE */
#ifdef	T_REQUEST
    {T_REQUEST, "req"},
    {T_REQUEST, "request"},
#endif	/* T_REQUEST */
#ifdef	T_RESOLVE
    {T_RESOLVE, "resolv"},
    {T_RESOLVE, "resolve"},
#endif	/* T_RESOLVE */
#ifdef	T_RESPONSE
    {T_RESPONSE, "response"},
#endif	/* T_RESPONSE */
#ifdef	T_RESTRICT
    {T_RESTRICT, "restrict"},
#endif	/* T_RESTRICT */
#ifdef	T_RETAIN
    {T_RETAIN, "retain"},
#endif	/* T_RETAIN */
#ifdef	T_RIP
    {T_RIP, "rip"},
#endif	/* T_RIP */
#ifdef	T_ROUTE
    {T_ROUTE, "route"},
#endif	/* T_ROUTE */
#ifdef	T_ROUTES
    {T_ROUTES, "routes"},
#endif	/* T_ROUTES */
#ifdef	T_ROUTING
    {T_ROUTING, "routing"},
#endif	/* T_ROUTING */
#ifdef	T_ROUTERDEADINTERVAL
    {T_ROUTERDEADINTERVAL, "router-dead-interval"},
    {T_ROUTERDEADINTERVAL, "routerdeadinterval"},
#endif	/* T_ROUTERDEADINTERVAL */
#ifdef	T_ROUTERDISCOVERY
    {T_ROUTERDISCOVERY, "router-discovery"},
    {T_ROUTERDISCOVERY, "routerdiscovery"},
#endif	/* T_ROUTERDISCOVERY */
#ifdef	T_ROUTERID
    {T_ROUTERID, "router-id"},
    {T_ROUTERID, "routerid"},
#endif	/* T_ROUTERID */
#ifdef	T_ROUTERS
    {T_ROUTERS, "routers"},
#endif	/* T_ROUTERS */
#ifdef	T_RXMITINTERVAL
    {T_RXMITINTERVAL, "retransmit-interval"},
    {T_RXMITINTERVAL, "retransmitinterval"},
    {T_RXMITINTERVAL, "rxmtinterval"},
#endif	/* T_RXMITINTERVAL */
#ifdef	T_SCANINTERVAL
    {T_SCANINTERVAL, "scan-interval"},
    {T_SCANINTERVAL, "scaninterval"},
#endif	/* T_SCANINTERVAL */
#ifdef	T_SEND
    {T_SEND, "send"},
#endif	/* T_SEND */
#ifdef	T_SENDBUF
    {T_SENDBUF, "send-buffer"},
    {T_SENDBUF, "sendbuf"},
    {T_SENDBUF, "sendbuffer"},
#endif	/* T_SENDBUF */
#ifdef	T_SET
    {T_SET,	"set"},
#endif	/* T_SET */
#ifdef	T_SETPREF
    {T_SETPREF,	"setpref"},
#endif	/* T_SETPREF */
#ifdef	T_SIMPLE
    {T_SIMPLE, "simple"},
    {T_SIMPLE, "simple-password"},
    {T_SIMPLE, "simplepassword"},
#endif	/* T_SIMPLE */
#ifdef	T_SIMPLEX
    {T_SIMPLEX, "simplex"},
#endif	/* T_SIMPLEX */
#ifdef	T_SIZE
    {T_SIZE, "size"},
#endif	/* T_SIZE */
#ifdef	T_SLSP
    {T_SLSP, "slsp"},
#endif	/* T_SLSP */
#ifdef	T_SNMP
    {T_SNMP, "snmp"},
#endif	/* T_SNMP */
#ifdef	T_SNPA
    {T_SNPA,	"snpa"},
#endif	/* T_SNPA */
#ifdef	T_SOURCEGATEWAYS
    {T_SOURCEGATEWAYS, "sourcegateways"},
#endif	/* T_SOURCEGATEWAYS */
#ifdef	T_SOURCENET
    {T_SOURCENET, "sourcenet"},
#endif	/* T_SOURCENET */
#ifdef	T_SPF
    {T_SPF, "spf"},
#endif	/* T_SPF */
#ifdef	T_STATE
    {T_STATE, "state"},
#endif	/* T_STATE */
#ifdef	T_STATIC
    {T_STATIC, "static"},
#endif	/* T_STATIC */
#ifdef	T_STRICTIFS
    {T_STRICTIFS, "strict-interfaces"},
    {T_STRICTIFS, "strict-intfs"},
    {T_STRICTIFS, "strictinterfaces"},
    {T_STRICTIFS, "strictintfs"},
#endif	/* T_STRICTIFS */
#ifdef	T_STUB
    {T_STUB, "stub"},
#endif	/* T_STUB */
#ifdef	T_STUBHOSTS
    {T_STUBHOSTS, "stub-hosts"},
    {T_STUBHOSTS, "stubhosts"},
#endif	/* T_STUBHOSTS */
#ifdef	T_SYMBOLS
    {T_SYMBOLS, "symbols"},
#endif	/* T_SYMBOLS */
#ifdef	T_SYSLOG
    {T_SYSLOG, "syslog"},
#endif	/* T_SYSLOG */
#ifdef	T_SYSTEMID
    {T_SYSTEMID,	"system-id"},
    {T_SYSTEMID,	"systemid"},
#endif	/* T_SYSTEMID */
#ifdef	T_TAG
    {T_TAG, "tag"},
#endif	/* T_TAG */
#ifdef	T_TASK
    {T_TASK, "task"},
#endif	/* T_TASK */
#ifdef	T_TEST
    {T_TEST, "test" },
#endif	/* T_TEST */
#ifdef	T_THRESHOLD
    {T_THRESHOLD, "threshold"},
#endif	/* T_THRESHOLD */
#ifdef	T_TIMER
    {T_TIMER, "timer"},
#endif	/* T_TIMER */
#ifdef	T_TRACEOPTIONS
    {T_TRACEOPTIONS, "traceoptions"},
#endif	/* T_TRACEOPTIONS */
#ifdef	T_TRANSITAREA
    {T_TRANSITAREA, "transit-area"},
    {T_TRANSITAREA, "transitarea"},
#endif	/* T_TRANSITAREA */
#ifdef	T_TRAP
    {T_TRAP, "trap"},
#endif	/* T_TRAP */
#ifdef	T_TROLL
    {T_TROLL,	"troll"},
#endif	/* T_TROLL */
#ifdef	T_TRUSTEDGATEWAYS
    {T_TRUSTEDGATEWAYS, "trustedgateways"},
#endif	/* T_TRUSTEDGATEWAYS */
#ifdef	T_TUNNEL
    {T_TUNNEL, "tunnel"},
#endif	/* T_TUNNEL */
#ifdef	T_TYPE
    {T_TYPE, "type"},
#endif	/* T_TYPE */
#ifdef	T_UNREACH
    {T_UNREACH, "unreach"},
#endif	/* T_UNREACH */
#ifdef	T_UPDATE
    {T_UPDATE, "update"},
#endif	/* T_UPDATE */
#ifdef	T_UPTO
    {T_UPTO,	"upto"},
#endif	/* T_UPTO */
#ifdef	T_VERSION
    {T_VERSION, "version"},
#endif	/* T_VERSION */
#ifdef	T_VIRTUALLINK
    {T_VIRTUALLINK, "virtual-link"},
    {T_VIRTUALLINK, "virtuallink"},
#endif	/* T_VIRTUALLINK */
#ifdef	T_WARNING
    {T_WARNING, "warning"},
#endif	/* T_WARNING */
/* End Sort */
#if	YYDEBUG != 0
    {T_YYDEBUG, "yydebug"},
    {T_YYQUIT, "yyquit"},
    {T_YYSTATE, "yystate"},
#endif	/* YYDEBUG != 0 */
    {0, NULL}
};


/*
 *	A string of alpha characters which is either a keyword or a host/network name.
 *	First do a binary search (Knuth 6.2.1) to lookup up a keyword, if it is not found
 *	assume it is a host/network name.  The proper thing to do is a REJECT, but flex
 *	won't optimize if we use a REJECT.
 */
bits *
parse_keyword __PF1(yytext, char *)
{
    register int c;
    register bits *i, *l, *u;

    l = keywords;
    u = &keywords[n_keywords - 1];
    do {
	i = (u - l) / 2 + l;
	c = strcasecmp(yytext, i->t_name);
	if (!c) {
	    return i;
	} else if (c < 0) {
	    u = i - 1;
	} else {
	    l = i + 1;
	}
    } while (u >= l);

    return (bits *) 0;
}


/*
 *	Look up a token name given it's ID in the keyword table.  There is no way to select
 *	between multiple keywords mapping to the same token once the table is sorted.
 */
bits *
parse_keyword_lookup __PF1(token, int)
{
    bits *p;

    for (p = keywords; p->t_name; p++) {
	if (token == p->t_bits) {
	    return p;
	}
    }

    return (bits *) 0;
}


/*
 *	A comparison routine used by the sorting routine that follows
 */
static int
parse_keyword_sort_compare __PF2(p1, const void_t,
				 p2, const void_t)
{
    return strcasecmp(((bits *) p1)->t_name,
		      ((bits *) p2)->t_name);
}


/*
 *	Quicksort the table of keywords to insure that they are in order, easier
 *	than trying to keep the source sorted and less prone to mistakes.
 *
 *	The sort is only done the first time we are called, i.e. when n_keyword is
 *	zero.
 */
static void
parse_keyword_sort __PF0(void)
{
    bits *p;

    if (!n_keywords) {
	/* Calculate size of table */
	for (p = keywords; p->t_name; p++, n_keywords++) ;

	qsort((caddr_t) keywords, n_keywords, sizeof(bits), parse_keyword_sort_compare);
    }
}


/*
 *	Front end for yyparse().  Exit if any errors
 */
int
parse_parse __PF1(file, const char *)
{
    int errors;
    static int first_parse = TRUE;

    if (first_parse) {
	parse_keyword_sort();		/* Sort the keyword table */

    }
    parse_directory = task_mem_strdup((task *) 0, task_path_start);

    sethostent(1);
    setnetent(1);

    if (parse_open(task_mem_strdup((task *) 0, file))) {
	/* Indicate problem opening file */
	
	errors = -1;
    } else {
	protos_seen = (proto_t) 0;
	parse_state = PS_INITIAL;
	switch (errors = yyparse()) {
	case 0:
	    errors = yynerrs;
	    break;

	default:
	    errors = yynerrs ? yynerrs : 1;
	}
    }

    endhostent();
    endnetent();

    if (parse_directory) {
	task_mem_free((task *) 0, parse_directory);
	parse_directory = (char *) 0;
    }
	
    if (errors) {
	int errs = ABS(errors);
	
	trace_log_tf(trace_global,
		     TRC_NL_BEFORE|TRC_NL_AFTER,
		     LOG_ERR,
		     ("parse_parse: %d parse error%s",
		      errs,
		      errs > 1 ? "s" : ""));
    }
    first_parse = FALSE;

    return errors;
}


/*
 *	Format a message indicating the current line number and return
 *	a pointer to it.
 */
char *
parse_where __PF0(void)
{
    static char where[LINE_MAX];

    if (parse_filename) {
	(void) sprintf(where, "%s:%d",
		       parse_filename,
		       yylineno);
    } else {
	(void) sprintf(where, "%d",
		       yylineno);
    }

    return where;
}


/*
 *	Limit check a number
 */
int
parse_limit_check __PF4(type, const char *,
			value, u_int,
			lower, u_int,
			upper, u_int)
{
    if ((value < lower) || ((upper != (u_int) -1) && (value > upper))) {
	(void) sprintf(parse_error, "invalid %s value at '%u' not in range %u to %u",
		       type,
		       value,
		       lower,
		       upper);
	return 1;
    }
    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("parse: %s %s: %u",
	      parse_where(),
	      type,
	      value));
    return 0;
}


/*
 *	Look up a string as a host or network name, returning it's IP address
 *	in normalized network byte order.
 */
sockaddr_un *
parse_addr_hostname __PF1(hname, char *)
{
    struct hostent *hostent;
    const char *errmsg = 0;

    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("parse_addr_hostname: resolving %s",
	      hname));

    hostent = gethostbyname(hname);
    if (hostent) {
	switch (hostent->h_addrtype) {
#ifdef	PROTO_INET
	case AF_INET:
	{
	    struct in_addr addr;
#ifdef	h_addr
	    if (hostent->h_addr_list[1]) {
		/* XXX - For a gateway we could use just the address on our network if there was only one */
		(void) sprintf(parse_error, "host has multiple addresses at '%s'",
			       hname);
		return (sockaddr_un *) 0;
	    }
#endif	/* h_addr */

	    bcopy(hostent->h_addr, (caddr_t) &addr.s_addr, (size_t) hostent->h_length);
	    return sockbuild_in(0, addr.s_addr);
	}
#endif	/* PROTO_INET */

	default:
	    /* XXX - Should we pretend it does not exist if it is not an INET name? */
	    (void) sprintf(parse_error, "unsupported host family at '%s'",
			   hname);
	    return (sockaddr_un *) 0;
	}
    } else {
#ifdef	HOST_NOT_FOUND
	if (h_errno && (h_errno < h_nerr)) {
	    errmsg = h_errlist[h_errno];
	} else {
	    errmsg = "Unknown host";
	}
#else	/* HOST_NOT_FOUND */
	errmsg = "Unknown host";
#endif	/* HOST_NOT_FOUND */
    }
    (void) sprintf(parse_error, "error resolving '%s': %s",
		   hname,
		   errmsg);
    return (sockaddr_un *) 0;
}


/*
 *	Look up a string as a host or network name, returning it's IP address
 *	in normalized network byte order.
 */
sockaddr_un *
parse_addr_netname __PF1(hname, char *)
{
    struct netent *netent;

    trace_tf(trace_global,
	     TR_PARSE,
	     0,
	     ("parse_addr_netname: resolving %s",
	      hname));

    netent = getnetbyname(hname);
    if (netent) {

	switch (netent->n_addrtype) {
#ifdef	PROTO_INET
	case AF_INET:
	{
	    struct in_addr network;

	    network.s_addr = netent->n_net;
	    if (network.s_addr) {
		while (!(network.s_addr & 0xff000000)) {
		    network.s_addr <<= 8;
		}
	    }
	    return sockbuild_in(0, network.s_addr);
	}
#endif	/* PROTO_INET */

	default:
	    (void) sprintf(parse_error, "unsupported network family at '%s'",
			   hname);
	    return (sockaddr_un *) 0;
	}
    } else {
	(void) sprintf(parse_error, "error resolving '%s': network unknown",
		       hname);
    }
    return (sockaddr_un *) 0;
}


/*
 *	Append an advlist to another advlist
 */
int
parse_adv_append __PF3(list, adv_entry **,
		       new, adv_entry *,
		       freeit, int)
{
    adv_entry *alo, *aln, *last = NULL;

    /* Add this network to the end of the list */
    if (*list) {
	ADV_LIST(*list, alo) {
	    if (!alo->adv_next) {
		last = alo;
	    }
	    /* Scan list for duplicates */
	    for (aln = new; aln; aln = aln->adv_next) {
		if (aln->adv_proto != alo->adv_proto
		    || (aln->adv_flag & ADVF_TYPE) != (alo->adv_flag & ADVF_TYPE)) {
		    continue;
		}
		switch (aln->adv_flag & ADVF_TYPE) {
		    case ADVFT_ANY:
			break;

		    case ADVFT_GW:
			if (aln->adv_gwp == alo->adv_gwp) {
			    (void) sprintf(parse_error, "duplicate gateway in list at '%A'",
					   aln->adv_gwp->gw_addr);
			    return TRUE;
			}
			break;

		    case ADVFT_IFN:
			if (aln->adv_ifn == alo->adv_ifae) {
			    (void) sprintf(parse_error, "duplicate interface in list at '%A'",
					   aln->adv_ifn->ifae_addr);
			    return TRUE;
			}
			break;

		    case ADVFT_IFAE:
			if (aln->adv_ifae == alo->adv_ifae) {
			    (void) sprintf(parse_error, "duplicate interface address in list at '%A'",
					   aln->adv_ifae->ifae_addr);
			    return TRUE;
			}
			break;

		    case ADVFT_AS:
			if (aln->adv_as == alo->adv_as) {
			    (void) sprintf(parse_error, "duplicate autonomous-system in list at '%u'",
					   aln->adv_as);
			    return TRUE;
			}
			break;

		    case ADVFT_DM:
			if (aln->adv_dm.dm_dest
			    && alo->adv_dm.dm_dest
			    && socktype(aln->adv_dm.dm_dest) != socktype(alo->adv_dm.dm_dest)) {
			    (void) sprintf(parse_error, "conflicting address family in dest/mask list at '%A mask %A'",
					   aln->adv_dm.dm_dest,
					   aln->adv_dm.dm_mask);
			    return TRUE;
			}
			if ((aln->adv_dm.dm_dest == alo->adv_dm.dm_dest
			     || sockaddrcmp(aln->adv_dm.dm_dest, alo->adv_dm.dm_dest))
			    && aln->adv_dm.dm_mask == alo->adv_dm.dm_mask) {
			    (void) sprintf(parse_error, "duplicate dest and mask in list at '%A mask %A'",
					   aln->adv_dm.dm_dest,
					   aln->adv_dm.dm_mask);
			    return TRUE;
			}
			break;

#ifdef	PROTO_ASPATHS
		    case ADVFT_TAG:
			if (aln->adv_tag == alo->adv_tag) {
			    (void) sprintf(parse_error, "duplicate tag in last at '%A'",
					   sockbuild_in(0, aln->adv_tag));
			    return TRUE;
			}
			break;
			
		    case ADVFT_ASPATH:
			if (aspath_adv_compare(alo->adv_aspath, aln->adv_aspath)) {
			    (void) sprintf(parse_error, "duplicate AS path pattern in list at '%s'",
					   aspath_adv_print(aln->adv_aspath));
			}
			break;
#endif	/* PROTO_ASPATHS */

		    case ADVFT_PS:
			if (PS_FUNC_VALID(aln, ps_compare) && PS_FUNC(aln, ps_compare)(alo->adv_ps, aln->adv_ps)) {
			    (void) sprintf(parse_error, "duplicate protocol specific data in list");
			    if (PS_FUNC_VALID(aln, ps_print)) {
				(void) sprintf(&parse_error[strlen(parse_error)],
					       "at '%s'",
					       PS_FUNC(aln, ps_print)(aln->adv_ps, TRUE));
			    }
			    return TRUE;
			}
			break;
		}
	    }
	} ADV_LIST_END(*list, alo) ;
	last->adv_next = new;
    } else {
	*list = new;
    }
    if (new) {
	/* XXX - bump the refcount, then free it? */
	ADV_LIST(new, aln) {
	    aln->adv_refcount++;
	} ADV_LIST_END(new, aln) ;
	if (freeit) {
	    adv_free_list(new);
	}
    }
    return FALSE;
}


/*
 *	Set a flag in the gw structure for each element in a list
 */
int
parse_gw_flag __PF3(list, adv_entry *,
		    proto, proto_t,
		    flag, flag_t)
{
    int n = 0;
    adv_entry *adv;

    ADV_LIST(list, adv) {
	BIT_SET(adv->adv_gwp->gw_flags, flag);
	adv->adv_gwp->gw_proto = proto;
	n++;
    } ADV_LIST_END(list, adv) ;
    return n;
}


/*
 *	Switch to a new state if it is a valid progression from
 *	the current state
 */
int
parse_new_state __PF1(state, int)
{
    static const bits states[] = {
        { PS_INITIAL,	"initial" },
        { PS_OPTIONS,	"options" },
        { PS_INTERFACE,	"interface" },
	{ PS_DEFINE,	"define" },
	{ PS_PROTO,	"protocol" },
	{ PS_ROUTE,	"route" },
	{ PS_CONTROL,	"control" },
	{ 0, NULL }
    } ;

    assert(state >= PS_MIN && state <= PS_MAX);

    if (state < parse_state) {
	(void) sprintf(parse_error, "statement out of order");
	return TRUE;
    } else if (state > parse_state) {
	trace_tf(trace_global,
		 TR_PARSE,
		 0,
		 ("parse_new_state: %s old %s new %s",
		  parse_where(),
		  trace_state(states, parse_state),
		  trace_state(states, state)));
	parse_state = state;
    }

    return FALSE;
}


int
parse_metric_check __PF2(proto, proto_t,
			 metric, pmet_t *)
{
    int rc = 1;
    
    struct limit {
	proto_t proto;
	metric_t low;
	metric_t high;
	const char *desc;
    };
    static struct limit limits[] = {
#ifdef	PROTO_RIP
        { RTPROTO_RIP, RIP_LIMIT_METRIC, "RIP metric" },
#endif	/* PROTO_RIP */
#ifdef	PROTO_HELLO
        { RTPROTO_HELLO, HELLO_LIMIT_DELAY, "HELLO delay" },
#endif	/* PROTO_HELLO */
#ifdef	PROTO_OSPF
        { RTPROTO_OSPF, OSPF_LIMIT_COST, "OSPF cost" },
#endif	/* PROTO_OSPF */
#ifdef	PROTO_OSPF
        { RTPROTO_OSPF_ASE, OSPF_LIMIT_METRIC, "OSPF metric" },
#endif	/* PROTO_OSPF */
#ifdef	PROTO_EGP
        { RTPROTO_EGP, EGP_LIMIT_DISTANCE, "EGP distance" },
#endif	/* PROTO_EGP */
#ifdef	PROTO_BGP
        { RTPROTO_BGP, BGP_LIMIT_METRIC, "BGP metric" },
#endif	/* PROTO_BGP */
#ifdef	PROTO_SLSP
	{ RTPROTO_SLSP, SLSP_LIMIT_COST, "SLSP metric" },
#endif	/* PROTO_SLSP */
#ifdef  PROTO_ISIS
        { RTPROTO_ISIS, ISIS_LIMIT_METRIC, "IS-IS metric" },
#endif  /* PROTO_ISIS */
#ifdef	PROTO_DVMRP
        { RTPROTO_DVMRP, DVMRP_LIMIT_METRIC, "DVMRP metric" },
#endif  /* PROTO_DVMRP */
        { RTPROTO_DIRECT, IF_LIMIT_METRIC, "interface metric" },
        { 0, 0, 0, NULL }
    }, *limit = limits;

    for (limit = limits; limit->proto; limit++) {
	if (limit->proto == proto) {
	    break;
	}
    }

    if (limit->proto) {
	switch (metric->state) {
	    case PARSE_METRICS_UNSET:
		(void) sprintf(parse_error, "parse_metric_check: %s metric not set",
			       trace_state(rt_proto_bits, proto));
		break;

	    case PARSE_METRICS_INFINITY:
		PARSE_METRIC_SET(metric, limit->high);
		/* Fall Thru */

	    case PARSE_METRICS_SET:
		rc = parse_limit_check(limit->desc,
				       (u_int) metric->metric,
				       (u_int) limit->low,
				       (u_int) limit->high);
		break;

	    default:
		(void) sprintf(parse_error, "parse_metric_set: %s metric in unknown state: %d",
			       trace_state(rt_proto_bits, proto),
			       metric->state);
	}
    } else {
	(void) sprintf(parse_error, "parse_metric_check: invalid protocol %s (%x)",
		       trace_state(rt_proto_bits, proto),
		       proto);
    }

    return rc;
}


/*
 *	Set metric for each element in list that does not have one
 */
adv_entry *
parse_adv_propagate_metric __PF4(list, adv_entry *,
				 proto, proto_t,
				 metric, pmet_t *,
				 advlist, adv_entry *)
{
    adv_entry *adv;

    ADV_LIST(list, adv) {
	adv->adv_proto = proto;
	if (parse_adv_append(&adv->adv_list, advlist, FALSE)) {
	    return (adv_entry *) 0;
	}
	if (PARSE_METRIC_ISRESTRICT(metric)) {
	    BIT_SET(adv->adv_flag, ADVF_NO);
	} else if (PARSE_METRIC_ISSET(metric)) {
	    BIT_SET(adv->adv_flag, ADVFOT_METRIC);
	    adv->adv_result.res_metric = metric->metric;
	}
    } ADV_LIST_END(list, adv) ;

    return list;
}


/*
 *	Set metric for each element in list that does not have one
 */
adv_entry *
parse_adv_propagate_preference __PF4(list, adv_entry *,
				     proto, proto_t,
				     preference, pmet_t *,
				     advlist, adv_entry *)
{
    adv_entry *adv;

    ADV_LIST(list, adv) {
	adv->adv_proto = proto;
	if (parse_adv_append(&adv->adv_list, advlist, FALSE)) {
	    return (adv_entry *) 0;
	}
	if (PARSE_METRIC_ISRESTRICT(preference)) {
	    BIT_SET(adv->adv_flag, ADVF_NO);
	} else if (PARSE_METRIC_ISSET(preference)) {
	    BIT_SET(adv->adv_flag, ADVFOT_METRIC);
	    adv->adv_result.res_preference = preference->metric;
	}
    } ADV_LIST_END(list, adv) ;

    return list;
}


/*
 *	Set config list for each element
 */
adv_entry *
parse_adv_propagate_config __PF3(list, adv_entry *,
				 config, config_list *,
				 proto, proto_t)
{
    adv_entry *adv;

    if (config) {
	ADV_LIST(list, adv) {
	    adv->adv_config = config;
	    adv->adv_proto = proto;
	    config->conflist_refcount++;
	} ADV_LIST_END(list, adv) ;
    }

    return list;
}


/*
 *	Set preference for each elmit in list that does not have one
 */
void
parse_adv_preference __PF3(list, adv_entry *,
			   proto, proto_t,
			   preference, pref_t)
{
    ADV_LIST(list, list) {
	if ((list->adv_flag & ADVFO_TYPE) == ADVFOT_NONE) {
	    list->adv_proto = proto;
	    BIT_SET(list->adv_flag, ADVFOT_PREFERENCE);
	    list->adv_result.res_preference = preference;
	}
    } ADV_LIST_END(list, list) ;
}


/*
 *	Return a pointer to a duplicate of this adv_entry
 */
adv_entry *
parse_adv_dup __PF1(old, adv_entry *)
{
    adv_entry *new = NULL, *adv, *root;

    root = NULL;

    ADV_LIST(old, old) {
	adv = adv_alloc(old->adv_flag, old->adv_proto);
	bcopy((caddr_t) old, (caddr_t) adv, sizeof(*adv));
	if (!root) {
	    root = adv;
	} else {
	    new->adv_next = adv;
	}
	if (adv->adv_list) {
	    adv->adv_list->adv_refcount++;
	}
	new = adv;
    } ADV_LIST_END(old, old) ;
    return root;
}


/*
 *	Lookup the entry in the list for the exterior protocol and append this list to it.
 */
int
parse_adv_as __PF2(advlist, adv_entry **,
		   adv, adv_entry *)
{
    adv_entry *list;

    ADV_LIST(*advlist, list) {
	if (adv->adv_as == list->adv_as) {
	    break;
	}
    } ADV_LIST_END(*advlist, list) ;

    if (!list) {
	list = adv_alloc(ADVFT_AS, adv->adv_proto);
	list->adv_as = adv->adv_as;
	if (parse_adv_append(advlist, list, TRUE)) {
	    return 0;
	}
    }
    if (parse_adv_append(&list->adv_list, adv, TRUE)) {
	return 0;
    }
    return 1;
}


/*
 *	Parse the config file options
 */
int
parse_args __PF2(argc, int,
		 argv, char **)
{
    int arg_n, err_flag = 0;
    char *arg, *cp, *ap;
    char seen[MAXHOSTNAMELENGTH];

    bzero(seen, MAXHOSTNAMELENGTH);

    for (arg_n = 1; arg_n < argc; arg_n++) {
	arg = argv[arg_n];
	if (*arg == '-') {
	    cp = arg + 1;
	    if (index(seen, (int) *cp)) {
		trace_log_tf(trace_global,
			     0,
			     LOG_ERR,
			     ("%s: duplicate switch: %s\n",
			      task_name,
			      arg));
		err_flag++;
		continue;
	    }
	    seen[strlen(seen)] = *cp;
	    switch (*cp++) {
	    case 'C':
		/* Test configuration */
		trace_nosyslog = TRACE_LOG_TRACE;
		task_newstate(TASKS_TEST|TASKS_NOSEND|TASKS_NODUMP, 0);
		BIT_SET(trace_global->tr_control, TRC_NOSTAMP);
		break;

	    case 'c':
		/* Test configuration */
		trace_nosyslog = TRACE_LOG_NONE;
		task_newstate(TASKS_TEST|TASKS_NOSEND, 0);
		trace_global->tr_flags = TR_GENERAL;
		break;

	    case 'n':
		/* Don't install in kernel */
		krt_install = FALSE;
		break;

	    case 'N':
		/* Don't daemonize */
		task_newstate(TASKS_NODAEMON, 0);
		break;
		
	    case 't':
		/* Set trace flags */
		if (*cp) {
		    if (!(trace_global->tr_flags = trace_args(cp))) {
			err_flag++;
		    }
		} else {
		    trace_global->tr_flags = TR_GENERAL;
		}
		break;

	    case 'f':
		/* Specify config file */
		ap = arg + 2;
		if (!*ap) {
		    if (((arg_n + 1) < argc) && (*argv[arg_n + 1] != '-')) {
			ap = argv[++arg_n];
		    }
		}
		if (!*ap) {
		    trace_log_tf(trace_global,
				 0,
				 LOG_ERR,
				 ("%s: missing argument for switch: %s\n",
				  task_progname,
				  arg));
		    err_flag++;
		    break;
		}
		if (task_config_file) {
		    task_mem_free((task *) 0, task_config_file);
		}
		task_config_file = task_mem_strdup((task *) 0, ap);
		break;

	    default:
		trace_log_tf(trace_global,
			     0,
			     LOG_ERR,
			     ("%s: invalid switch: %s\n",
			      task_progname,
			      arg));
		err_flag++;
	    }
	} else if (!trace_global->tr_file->trf_file) {
	    trace_file_free(trace_global->tr_file);
	    trace_global->tr_file = trace_file_locate(task_mem_strdup((task *) 0, arg),
						      (off_t) 0,
						      (u_int) 0,
						      (flag_t) 0);
	    trace_on(trace_global->tr_file);
	} else {
	    trace_log_tf(trace_global,
			 0,
			 LOG_ERR,
			 ("%s: extraneous information on command line: %s\n",
			  task_progname,
			  arg));
	    err_flag++;
	}
    }

    if (err_flag) {
	trace_log_tf(trace_global,
		     0,
		     LOG_ERR,
		     ("Usage: %s [-c] [-C] [-n] [-N] [-t[flags]] [-f config-file] [trace-file]\n",
		      task_progname));
    }
    return err_flag;
}


/*
 * ------------------------------------------------------------------------
 * 
 * 	GateD, Release 3
 * 
 * 	Copyright (c) 1990,1991,1992,1993 by Cornell University
 * 	    All rights reserved.
 * 
 * 	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
 * 	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * 	LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * 	AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 	Royalty-free licenses to redistribute GateD Release
 * 	3 in whole or in part may be obtained by writing to:
 * 
 * 	    GateDaemon Project
 * 	    Information Technologies/Network Resources
 * 	    200 CCC
 * 	    Cornell University
 * 	    Ithaca, NY  14853-2601  USA
 * 
 * 	GateD is based on Kirton's EGP, UC Berkeley's routing
 * 	daemon	 (routed), and DCN's HELLO routing Protocol.
 * 	Development of GateD has been supported in part by the
 * 	National Science Foundation.
 * 
 * 	Please forward bug fixes, enhancements and questions to the
 * 	gated mailing list: gated-people@gated.cornell.edu.
 * 
 * 	Authors:
 * 
 * 		Jeffrey C Honig <jch@gated.cornell.edu>
 * 		Scott W Brim <swb@gated.cornell.edu>
 * 
 * ------------------------------------------------------------------------
 * 
 *       Portions of this software may fall under the following
 *       copyrights:
 * 
 * 	Copyright (c) 1988 Regents of the University of California.
 * 	All rights reserved.
 * 
 * 	Redistribution and use in source and binary forms are
 * 	permitted provided that the above copyright notice and
 * 	this paragraph are duplicated in all such forms and that
 * 	any documentation, advertising materials, and other
 * 	materials related to such distribution and use
 * 	acknowledge that the software was developed by the
 * 	University of California, Berkeley.  The name of the
 * 	University may not be used to endorse or promote
 * 	products derived from this software without specific
 * 	prior written permission.  THIS SOFTWARE IS PROVIDED
 * 	``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * 	INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * 	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
